import time
import datetime
import pprint
from picamera2 import Picamera2
from libcamera import controls


# Fast Exposure: make next exposure while the processing of the actual one is still ongoing
DoFastExposure = True


# enable picamera2 debug logging
#Picamera2.set_logging(Picamera2.DEBUG)


# list of available cameras:
cameras = Picamera2.global_camera_info()
print()
print("cameras:")
pprint.pprint(cameras)
# [{'Id': '/base/soc/i2c0mux/i2c@1/imx477@1a',
#   'Location': 2,
#   'Model': 'imx477',
#   'Rotation': 180}]
#
# CSI camaeras are listed ahead of USB cameras. Order of USB cameras can change!


# using first camera in list
picam2 = Picamera2(0)


# camera properties
print()
print("camera properties:")
pprint.pprint(picam2.camera_properties)
# {'ColorFilterArrangement': 0,
#  'Location': 2,
#  'Model': 'imx477',
#  'PixelArrayActiveAreas': (8, 16, 4056, 3040),
#  'PixelArraySize': (4056, 3040),
#  'Rotation': 180,
#  'ScalerCropMaximum': (0, 0, 0, 0),
#  'UnitCellSize': (1550, 1550)}


# we want to do still exposures
# For Fast Exposure we need a queue with more than one buffer. 
config = picam2.create_still_configuration(queue=DoFastExposure, buffer_count=2 if DoFastExposure else 1)
print()
print("default config:")
pprint.pprint(config)
# {'buffer_count': 2,
#  'colour_space': <libcamera.ColorSpace 'sYCC'>,
#  'controls': {'FrameDurationLimits': (100, 1000000000),
#               'NoiseReductionMode': <NoiseReductionModeEnum.HighQuality: 2>},
#  'display': None,
#  'encode': None,
#  'lores': None,
#  'main': {'format': 'BGR888', 'size': (4056, 3040)},
#  'queue': True,
#  'raw': None,
#  'transform': <libcamera.Transform 'identity'>,
#  'use_case': 'still'}


# Libcamarea can provide 3 streams: raw, main and lores.
# Main stream is always generated. Raw stream data is taken before any signal processing in the image signal processor (ISP).
# Processed images have very low value for astro-photography. In fact we can ignore the main stream nd focus on the raw stream.
# The main stream cannot be disabled. Therefore we configure the main stream to a low resolution to save memory (very important
# on Raspberry Pi Zero!)
config["main"]["size"] = (504, 380)


# Raw camera modes allow only a small set of size, bit depth and format combinations:
print()
print("raw sensor modes:")
pprint.pprint(picam2.sensor_modes)
# [{'bit_depth': 10,
#   'crop_limits': (696, 528, 2664, 1980),
#   'exposure_limits': (31, 667234896, None),
#   'format': SRGGB10_CSI2P,
#   'fps': 120.05,
#   'size': (1332, 990),
#   'unpacked': 'SRGGB10'},
#  {'bit_depth': 12,
#   'crop_limits': (0, 440, 4056, 2160),
#   'exposure_limits': (60, 674181621, None),
#   'format': SRGGB12_CSI2P,
#   'fps': 50.03,
#   'size': (2028, 1080),
#   'unpacked': 'SRGGB12'},
#  {'bit_depth': 12,
#   'crop_limits': (0, 0, 4056, 3040),
#   'exposure_limits': (60, 674181621, None),
#   'format': SRGGB12_CSI2P,
#   'fps': 40.01,
#   'size': (2028, 1520),
#   'unpacked': 'SRGGB12'},
#  {'bit_depth': 12,
#   'crop_limits': (0, 0, 4056, 3040),
#   'exposure_limits': (114, 694422939, None),
#   'format': SRGGB12_CSI2P,
#   'fps': 10.0,
#   'size': (4056, 3040),
#   'unpacked': 'SRGGB12'}]
#
# Taken from picamera2 documentation:
# This gives us the exact sensor modes that we can request, with the following information for each mode:
# • bit_depth - the number of bits in each pixel sample.
# • crop_limits - this tells us the exact field of view of this mode within the full resolution sensor output. In the example
# above, only the final two modes will give us the full field of view.
# • exposure_limits - the maximum and minimum exposure values (in microseconds) permitted in this mode.
# • format - the packed sensor format. This can be passed as the "format" when requesting the raw stream.
# • fps - the maximum framerate supported by this mode.
# • size - the resolution of the sensor output. This value can be passed as the "size" when requesting the raw stream.
# • unpacked - use this in place of the earlier format in the raw stream request if unpacked raw images are required 
#
# Raw stream formats are "S" followed by Bayer pattern (except monochrome sensors which have only one channel "R") and number of bits.
# "_CSI2P": pixel data will be packed tightly in memory, so that four ten-bit pixel values will be stored in every five bytes,
#           or two twelve-bit pixel values in every three bytes.
# When _CSI2P is absent, it means the pixels will each be unpacked into a 16-bit word (or eight-bit pixels into a single byte).
# This uses more memory but can be useful for applications that want easy access to the raw pixel data.
# We need unpacked pixel to send them as FITS.
#
# Not all cameras supported by libcamera will provide a raw mode!
#
# If you want to choose the raw mode yourself, only the "format" and "size" fields have any effect but the other fields will be ignored:
config["raw"] = {"size": (4056, 3040), "format": 'SRGGB12'}


# optimize (allign) configuration: small changes to some main stream configurations (for instance: size) will fit better to hardware
picam2.align_configuration(config)

# set still configuration
picam2.configure(config)

print()
print("used config:")
pprint.pprint(picam2.camera_config)
# {'buffer_count': 2,
#  'colour_space': <libcamera.ColorSpace 'sYCC'>,
#  'controls': {'FrameDurationLimits': (100, 1000000000),
#               'NoiseReductionMode': <NoiseReductionModeEnum.HighQuality: 2>},
#  'display': None,
#  'encode': None,
#  'lores': None,
#  'main': {'format': 'BGR888',
#           'framesize': 547200,
#           'size': (480, 380),
#           'stride': 1440},
#  'queue': True,
#  'raw': {'format': 'SBGGR12',
#          'framesize': 24709120,
#          'size': (4056, 3040),
#          'stride': 8128},
#  'transform': <libcamera.Transform 'identity'>,
#  'use_case': 'still'}


# runtime controls: can be changed when camera runs
print()
print("runtime controls:")
pprint.pprint(picam2.controls.make_dict())
# {'FrameDurationLimits': (100, 1000000000),
# 'NoiseReductionMode': <NoiseReductionModeEnum.HighQuality: 2>}


# The default runtime controls are very rudimentary and we need to set more.
# We want to disable exposure regulations because they increae time to get frames and prevent later frame stacking.
# The only useful controls for raw frames are AnalogGain and ExposureTime.
min_exp, max_exp, default_exp = picam2.camera_controls["ExposureTime"]
print(f'ExposureTime: min {min_exp}, max {max_exp}, default {default_exp}')
# ExposureTime: min 114, max 694422939, default None
min_again, max_again, default_again = picam2.camera_controls["AnalogueGain"]
print(f'AnalogGain: min {min_again}, max {max_again}, default {default_again}')
# AnalogGain: min 1.0, max 22.2608699798584, default None
picam2.set_controls(
    {
        # controls for main frame: disable all regulations
        "AeEnable": False,  # AEC/AGC algorithm
        "NoiseReductionMode": controls.draft.NoiseReductionModeEnum.Off,  # disable noise reduction in main frame because it eats stars
        "AwbEnable": False,  # disable automatic white balance algorithm
        # controls for raw and main frames
        "AnalogueGain": max_again,  # max AnalogGain
        "ExposureTime": int(1*1e6),  # exposure time in us; needs to be integer!
    }
)

print()
print("used runtime controls:")
pprint.pprint(picam2.controls.make_dict())
# {'AeEnable': False,
#  'AnalogueGain': 22.2608699798584,
#  'AwbEnable': False,
#  'ExposureTime': 1000000,
#  'FrameDurationLimits': (100, 1000000000),
#  'NoiseReductionMode': <NoiseReductionModeEnum.Off: 0>}


# Picamera2 always needs a preview window (because it drives the libcamera system, forwards the image to the application and recycles the buffers).
# The NULL preview shows nothing.
# NULL is automatically started with picam2.start() when no other preview is already running.
picam2.start()


# exposure and return 2 streams + metadata
# array_raw, array_main and metadata are from the same exposure
print()
print("exposure:")
(array_main, array_raw), metadata = picam2.capture_arrays(["main", "raw"])
print(f'main array: {type(array_main)}, shape={array_main.shape}, dtype={array_main.dtype}')
# main array: <class 'numpy.ndarray'>, shape=(380, 480, 3), dtype=uint8
# 380 x 480 pixel ith each 3 channels (BGR888)
print(f'raw array: {type(array_raw)}, shape={array_raw.shape}, dtype={array_raw.dtype}')
# raw array: <class 'numpy.ndarray'>, shape=(3040, 8128), dtype=uint8
# This is the raw Bayer data with 2 byte per pixel! Strange that the array is of type uint8 and has the 2 bytes/pixel separated.
# TODO: endieness of the 2 bytes, are the 12bit values left or right aligned? 
#       will that work: array_raw_16bit = array_raw.view(np.uint16) * 16  # convert 12 bit values to 16 bit
pprint.pprint(metadata)
# {'AeLocked': False,
#  'AnalogueGain': 22.2608699798584,
#  'ColourCorrectionMatrix': (2.1575498580932617,
#                             -1.1039637327194214,
#                             -0.05359601974487305,
#                             -0.3457588851451874,
#                             1.6122934818267822,
#                             -0.2665446400642395,
#                             -0.02813815139234066,
#                             -0.7009111046791077,
#                             1.7290493249893188),
#  'ColourGains': (2.510904550552368, 1.656680941581726),
#  'ColourTemperature': 4000,
#  'DigitalGain': 1.0,
#  'ExposureTime': 1000000,
#  'FrameDuration': 1000628,
#  'Lux': 400.0,
#  'ScalerCrop': (108, 0, 3840, 3040),
#  'SensorBlackLevels': (4096, 4096, 4096, 4096),
#  'SensorTemperature': 22.0,
#  'SensorTimestamp': 5318556086000}


# Stop camera before changing controls. Otherwise it will take some frames until the new setting becomes active. 
# Without stop/start I saw up to 4 frames still having the old exposure time!
picam2.stop()
# change exposure time
picam2.set_controls(
    {
        "ExposureTime": int(5*1e6),  # exposure time in us; needs to be integer!
    }
)
picam2.start()


print()
StartTime = datetime.datetime.now()
for i in range(10):
    #(array_main, array_raw), metadata = picam2.capture_arrays(["main", "raw"])
    (array_raw), metadata = picam2.capture_arrays(["raw"])
    print(f'exposure {i}: needed {datetime.datetime.now() - StartTime}, ExposureTime={metadata["ExposureTime"]}, FrameDuration={metadata["FrameDuration"]}')
    StartTime = datetime.datetime.now()
    # assume 2sec processing and transmission of frames
    time.sleep(2.0)
# exposure 0: needed 0:00:05.940195, ExposureTime=4999942, FrameDuration=5000571
# exposure 1: needed 0:00:04.995752, ExposureTime=4999942, FrameDuration=5000571
# exposure 2: needed 0:00:05.083414, ExposureTime=4999942, FrameDuration=5000571
# exposure 3: needed 0:00:04.918735, ExposureTime=4999942, FrameDuration=5000571
# exposure 4: needed 0:00:04.988870, ExposureTime=4999942, FrameDuration=5000571
# exposure 5: needed 0:00:05.024130, ExposureTime=4999942, FrameDuration=5000571
# exposure 6: needed 0:00:04.983263, ExposureTime=4999942, FrameDuration=5000571
# exposure 7: needed 0:00:04.964785, ExposureTime=4999942, FrameDuration=5000571
# exposure 8: needed 0:00:05.058017, ExposureTime=4999942, FrameDuration=5000571
# exposure 9: needed 0:00:04.970675, ExposureTime=4999942, FrameDuration=5000571
#
# Time between frames is about 5 seconds. That means exposure of next frame runs in parallel with
# the processing (here simulated with time.sleep(2.0)).


# Can we do extreme long term exposures? 
# picam2.camera_controls["ExposureTime"] would allow up to 694 seconds exposure. 
# But HQ camera documentation says 670.74 seconds (https://www.raspberrypi.com/documentation/accessories/camera.html).
# On other pages I saw even lower values for maximum exposure time (~280 second).
print()
picam2.stop()
# change exposure time
picam2.set_controls(
    {
        "ExposureTime": int(240*1e6),  # exposure time in us; needs to be integer!
    }
)
picam2.start()
StartTime = datetime.datetime.now()
(array_raw), metadata = picam2.capture_arrays(["raw"])
print(f'long term exposure: needed {datetime.datetime.now() - StartTime}, ExposureTime={metadata["ExposureTime"]}, FrameDuration={metadata["FrameDuration"]}')
# When using ExposureTime 690 seconds (longer than HQ documentation allows!) I get a 72 second exposure only.
# But when setting 240 second exposure time I get:
# long term exposure: needed 0:04:00.927936, ExposureTime=239991370, FrameDuration=239992000


# The camera already started the next long term exposure. Can we break this with stop/start or do we need to wait until the exposure ends?
print()
StartTime = datetime.datetime.now()
picam2.stop()
# change exposure time
picam2.set_controls(
    {
        "ExposureTime": int(1*1e6),  # exposure time in us; needs to be integer!
    }
)
picam2.start()
print(f'time needed to stop long term exposure: {datetime.datetime.now() - StartTime}')
# time needed to stop long term exposure: 0:03:57.792661
# To stop a running exposure we need to wait until the exposure finishes. That's not convenient when you want to switch back from
# very long exposure time to short.
# TODO: find way to stop running exposure immediately

StartTime = datetime.datetime.now()
(array_raw), metadata = picam2.capture_arrays(["raw"])
print(f'exposure: needed {datetime.datetime.now() - StartTime}, ExposureTime={metadata["ExposureTime"]}, FrameDuration={metadata["FrameDuration"]}')
# exposure: needed 0:00:01.937229, ExposureTime=1000000, FrameDuration=1000628


# probably not needed to call stop at end of program (will happen automatically)
picam2.stop()
